// @(#) ENIB/LI2 (c) - C++ Utilities
//--------------------------------------------------------------------------//
// file        : tcpUdp.h                                                   //
// description : header file for the ENIB/LI2 TCP and UDP communication     //
// author(s)   : F. Harrouet                                                //
// copyright   : Laboratoire d'Informatique Industrielle (LI2)              //
//               Ecole Nationale d'Ingenieurs de Brest (ENIB)               //
//               CP 30815 - 29608 Brest Cedex - FRANCE                      //
//               Tel: (33) 02.98.05.66.31 , Fax: (33) 02.98.05.66.29        //
//               e-mail: li2@enib.fr                                        //
//--------------------------------------------------------------------------//
// Copyright (c) 1996-2000                                                  //
// Laboratoire d'Informatique Industrielle (LI2)                            //
// Ecole Nationale d'Ingenieurs de Brest (ENIB)                             //
//                                                                          //
// Permission to use, copy, modify, distribute and sell this software       //
// and its documentation for any purpose is hereby granted without fee,     //
// provided that the above copyright notice appear in all copies and        //
// that both that copyright notice and this permission notice appear        //
// in supporting documentation.                                             //
// ENIB/LI2 makes no representations about the suitability of this          //
// software for any purpose.                                                //
// It is provided "as is" without express or implied warranty.              //
//--------------------------------------------------------------------------//

#ifndef TCPUDP_H
#define TCPUDP_H

#include "Utility/strings.h"
#include "Utility/smplThrd.h"
#include "Utility/stlMap.h"
#include "Utility/stlList.h"

#ifndef __MSDOS__
#include <netinet/in.h>
#else
#include <winsock.h>
#endif

//--------------------------------------------------------------------------//
// class SocketElem (abstract)                                              //
//--------------------------------------------------------------------------//

class SocketElem : public BlockAllocatedObject
{
public:
// Static Methods
  static void relaxSockets(void);
  static void * allocBytes(unsigned long nBytes);
  static void freeBytes(void * bytes);
  static Strings getHostName(void);
  static Strings getIPAddress(const Strings & hostName);
// Allocators
// Inspectors
  bool fail(void) const;
// Modifiers
protected:
// Private Methods
  SocketElem(void);
  virtual ~SocketElem(void);
  static unsigned long _getIPAddress(const Strings & hostName);
  static void _printAddress(Strings & result,unsigned long ipAddress);
  static void _closeSocket(int & socket);
// Attributes
  int _socket;
  typedef StlMap<Strings,unsigned long> IPMap;
  static IPMap _ipMap;
  static SimpleVector<int> _sockets;
  static bool _firstTime;
};

//--------------------------------------------------------------------------//
// class TCPListener                                                        //
//--------------------------------------------------------------------------//

class TCPConnection;

class TCPListener : public SocketElem
{
public:
// Allocators
  TCPListener(int portNumber);
  virtual ~TCPListener(void);
// Inspectors
  int getNbPending(void) const;
  int getPortNumber(void) const;
  const Strings & getIPAddress(void) const;
// Modifiers
  TCPConnection * nextConnection(void);
protected:
// Private methods
  static void * _acceptTask(void * data);
// Attributes
  SimpleThread * _acceptThread;
  SimpleThreadMutex _pendingMutex;
  typedef StlList<StlPair<int,struct sockaddr_in> > PendingList;
  PendingList _pending;
  int _portNumber;
  Strings _ipAddress;
};

//--------------------------------------------------------------------------//
// class TCPConnection                                                      //
//--------------------------------------------------------------------------//

class TCPConnection : public SocketElem
{
public:
// Allocators
  TCPConnection(int portNumber,const Strings & host);
  virtual ~TCPConnection(void);
// Inspectors
  int getLocalPortNumber(void) const;
  const Strings & getLocalIPAddress(void) const;
  int getDistantPortNumber(void) const;
  const Strings & getDistantIPAddress(void) const;
// Modifiers
  bool readBytes(void * & result,unsigned long & nBytes);
  bool read(Strings & result);
  void writeBytes(const void * bytes,unsigned long nBytes);
  void write(const Strings & str);
protected:
// Private methods
  friend class TCPListener;
  TCPConnection(int socket,struct sockaddr_in & socketAddress);
  static void * _receiveTask(void * data);
  bool _readNBytes(void * buffer,unsigned long n);
// Attributes
  typedef StlList<StlPair<unsigned long,void *> > BufferList;
  BufferList _received;
  SimpleThreadMutex _receiveMutex;
  SimpleThread * _receiveThread;
  int _localPortNumber,_distantPortNumber;
  Strings _localIPAddress,_distantIPAddress;
};

//--------------------------------------------------------------------------//
// class UDPTransmitter                                                     //
//--------------------------------------------------------------------------//

class UDPTransmitter : public SocketElem
{
public:
// Allocators
  UDPTransmitter(int portNumber);
  virtual ~UDPTransmitter(void);
// Inspectors
  int getPortNumber(void) const;
  const Strings & getIPAddress(void) const;
// Modifiers
  bool readBytes(void * & result,unsigned long & nBytes);
  bool readBytesFrom(void * & result,unsigned long & nBytes,
                     int & portNumber,Strings & ipAddress);
  bool read(Strings & result);
  bool readFrom(Strings & result,
                int & portNumber,Strings & ipAddress);
  void writeBytesTo(const void * bytes,unsigned long nBytes,
                    int portNumber,const Strings & host);
  void writeTo(const Strings & str,
               int portNumber,const Strings & host);
  void broadcastBytes(const void * bytes,unsigned long nBytes);
  void broadcast(const Strings & str);
protected:
// Private methods
  static void * _receiveTask(void * data);
// Attributes
  struct Buffer
    {
    unsigned long nBytes;
    void * buffer;
    unsigned long ipAddress;
    int portNumber;
    };
  typedef StlList<Buffer> BufferList;
  BufferList _received;
  SimpleThreadMutex _receiveMutex;
  SimpleThread * _receiveThread;
  char * _receptionBuffer;
  int _portNumber;
  Strings _ipAddress;
  struct sockaddr_in _socketAddress;
  static const int _receptionSize;
};

#endif // TCPUDP_H

