// @(#) ENIB/LI2 (c) - C++ Utilities
//--------------------------------------------------------------------------//
// file        : blockAlc.cpp                                               //
// description : source file for the ENIB/LI2 class BlockAllocator          //
// author(s)   : F. Harrouet                                                //
// copyright   : Laboratoire d'Informatique Industrielle (LI2)              //
//               Ecole Nationale d'Ingenieurs de Brest (ENIB)               //
//               CP 30815 - 29608 Brest Cedex - FRANCE                      //
//               Tel: (33) 02.98.05.66.31 , Fax: (33) 02.98.05.66.29        //
//               e-mail: li2@enib.fr                                        //
//--------------------------------------------------------------------------//
// Copyright (c) 1996-2000                                                  //
// Laboratoire d'Informatique Industrielle (LI2)                            //
// Ecole Nationale d'Ingenieurs de Brest (ENIB)                             //
//                                                                          //
// Permission to use, copy, modify, distribute and sell this software       //
// and its documentation for any purpose is hereby granted without fee,     //
// provided that the above copyright notice appear in all copies and        //
// that both that copyright notice and this permission notice appear        //
// in supporting documentation.                                             //
// ENIB/LI2 makes no representations about the suitability of this          //
// software for any purpose.                                                //
// It is provided "as is" without express or implied warranty.              //
//--------------------------------------------------------------------------//

#include "Utility/blockAlc.h"
#include "Utility/smplThrd.h"

#include <string.h>
#include <stdio.h>

//--------------------------------------------------------------------------//
// class BlockAllocator                                                     //
//--------------------------------------------------------------------------//

class AllocMutex : public SimpleThreadMutex // internal use
{
protected:
  static bool _initDone;
public:
  inline AllocMutex(BlockAllocator * alloc)
          : SimpleThreadMutex(alloc)
         {
         AllocMutex::_initDone=true;
         }
  inline virtual ~AllocMutex(void)
         {
         AllocMutex::_initDone=false;
         }
  inline void lock(void)
         {
         if(AllocMutex::_initDone) SimpleThreadMutex::lock();
         // else cerr << "AllocMutex::lock() init not done !" << endl;
         }
  inline void unlock(void)
         {
         if(AllocMutex::_initDone) SimpleThreadMutex::unlock();
         // else cerr << "AllocMutex::unlock() init not done !" << endl;
         }
};
bool AllocMutex::_initDone=false;
static AllocMutex mutex((BlockAllocator *)0);

int BlockAllocator::_nbCreated=0;

const size_t BlockAllocator::_sizeOffset=8; // 8 bytes seems to be a good
                                            // memory alignement
void * BlockAllocator::_blockList[1024];
                                           // this static member ensures 
BlockAllocator BlockAllocator::_allocator; // initialisation at the
                                           // application start and clearing
                                           // at the end

//-------- Private utilities -----------------------------------------------//
inline size_t BlockAllocator::_maxBlockSize(void) // static
{
return(32768);  // 1024 * 32
}

inline size_t BlockAllocator::_blockIndex(size_t size) // static
{
return(((--size)>>5)); // step = 32 bytes
}

inline size_t BlockAllocator::_blockSize(size_t size) // static
{
return((size+31)&~31); // step = 32 bytes
}

//-------- Allocators ------------------------------------------------------//
BlockAllocator::BlockAllocator(void)
{
if(!_nbCreated++)
  {
  for(size_t i=1024;i--;)
    _blockList[i]=(void *)0;
  }
}

BlockAllocator::~BlockAllocator(void)
{
if(!--_nbCreated)
  {
  void * block;
  for(size_t i=1024;i--;)
    {
    while(_blockList[i])
      {
      block=_blockList[i];
      _blockList[i]=*((void **)block);
      ::free(block);
      }
    }
  }
}

//-------- Memory management -----------------------------------------------//

void * BlockAllocator::allocate(size_t size) // static
{
void * block;
mutex.lock();
size+=_sizeOffset; // the size will be stored in the begining of the block
if(size>_maxBlockSize())
  {
  block=::malloc(size);
  }
else
  {
  size_t index=_blockIndex(size);
  block=_blockList[index];
  if(block)
    {
    _blockList[index]=*((void **)block); // next block ready
    }
  else
    {
    block=::malloc(_blockSize(size));
    }
  }
*((size_t *)block)=size; // store the size in the block
mutex.unlock();
return(((char *)block)+_sizeOffset); // give the next address
}

void BlockAllocator::free(void * block) // static
{
mutex.lock();
block=((char *)block)-_sizeOffset; // size stored at the previous address
size_t size=*((size_t *)block);
if(size>_maxBlockSize())
  {
  ::free(block);
  }
else
  {
  size_t index=_blockIndex(size);
  *((void **)block)=_blockList[index]; // save next block
  _blockList[index]=block;
  }
mutex.unlock();
}

void * BlockAllocator::realloc(void * block,size_t newSize) // static
{
void * newBlock;
mutex.lock();
block=((char *)block)-_sizeOffset; // size stored at the previous address
size_t oldSize=*((size_t *)block);
newSize+=_sizeOffset; // the size will be stored in the begining of the block
if(oldSize>_maxBlockSize())
  {
  if(newSize>_maxBlockSize())
    {
    newBlock=::realloc(block,newSize);
    }
  else
    {
    size_t index=_blockIndex(newSize);
    newBlock=_blockList[index];
    if(newBlock)
      {
      _blockList[index]=*((void **)newBlock); // next block ready
      }
    else
      {
      newBlock=::malloc(_blockSize(newSize));
      }
    ::memcpy(newBlock,block,newSize);
    ::free(block);
    }
  }
else
  {
  if(newSize>_maxBlockSize())
    {
    newBlock=::malloc(newSize);
    ::memcpy(newBlock,block,oldSize);
    size_t index=_blockIndex(oldSize);
    *((void **)block)=_blockList[index]; // save next block
    _blockList[index]=block;
    }
  else
    {
    size_t newBlockSize=_blockSize(newSize);
    size_t oldBlockSize=_blockSize(oldSize);
    if(oldBlockSize==newBlockSize)
      {
      newBlock=block;
      }
    else
      {
      size_t index=_blockIndex(newSize);
      newBlock=_blockList[index];
      if(newBlock)
        {
        _blockList[index]=*((void **)newBlock); // next block ready
        }
      else
        {
        newBlock=::malloc(_blockSize(newSize));
        }
      ::memcpy(newBlock,block,newSize<oldSize ? newSize : oldSize);
      index=_blockIndex(oldSize);
      *((void **)block)=_blockList[index]; // save next block
      _blockList[index]=block;
      }
    }
  }
*((size_t *)newBlock)=newSize; // store the size in the block
mutex.unlock();
return(((char *)newBlock)+_sizeOffset); // give the next address
}

void BlockAllocator::lock(void) // static
{
mutex.lock();
}

void BlockAllocator::unlock(void) // static
{
mutex.unlock();
}

