/*----------------------------------------------------------------------------
  tpC_graph.cpp
----------------------------------------------------------------------------*/

#include "tpC_graph.h"
#include "guiTrans.h"
#include "Utility/system.h"

#include <math.h>

#include <iostream>
using namespace std;

/*-------- Data types ------------------------------------------------------*/

typedef struct _Object2D
{
GuiShape2D * _shape;
bool _isImage;
} Object2D;

/*-------- Static data -----------------------------------------------------*/

static void * _userData=(void *)0;

static GuiWindow * _mainWin=(GuiWindow *)0;

static GuiViewer2D * _viewer=(GuiViewer2D *)0;

/*-------- Hidden Callbacks ------------------------------------------------*/

static
void
_keyCB(GuiShape2D * /* shape */,
       const Strings & key,
       void * userData)
{
graphic_keyPressCallback((Object2D *)userData,key,_userData);
}

static
void
_motionCB(GuiShape2D * /* shape */,
          double dx,
          double dy,
          void * userData)
{
graphic_mouseDragCallback((Object2D *)userData,dx,dy,_userData);
}

static
void
_viewerKeyCB(GuiViewer2D * /* viewer */,
             const Strings & key,
             void * /* userData */)
{
graphic_keyPressCallback((Object2D *)0,key,_userData);
}

static
void
_viewerResizeCB(GuiElement * calling,
                void * viewer)
{
GuiWindow * win=(GuiWindow *)calling;
GuiViewer2D * v2d=(GuiViewer2D *)viewer;
v2d->setX(0);
v2d->setY(0);
v2d->setWidth(win->getInWidth());
v2d->setHeight(win->getInHeight());
}

static
void
_viewerCloseCB(GuiElement *,
               void *)
{
exit(0);
}

/*-------- Construction/Destruction ----------------------------------------*/

Object2D *
Object2D_new(void)
{
Object2D * obj2d=new Object2D;
obj2d->_shape=new GuiPoint(0.0,0.0,0.0,"white");
obj2d->_shape->setKeyCallback(_keyCB,obj2d);
obj2d->_shape->setMotionCallback(_motionCB,obj2d);
obj2d->_isImage=false;
return(obj2d);
}

void
Object2D_delete(Object2D * obj2d)
{
delete obj2d->_shape;
delete obj2d;
}

/*-------- Location -------------------------------------------------------*/

void
Object2D_setLocation(Object2D * obj2d,
                     double x,
                     double y,
                     double theta)
{
obj2d->_shape->move(x,y,theta);
}

void
Object2D_getLocation(const Object2D * obj2d,
                     double * xOut,
                     double * yOut,
                     double * thetaOut)
{
*xOut=obj2d->_shape->getX();
*yOut=obj2d->_shape->getY();
*thetaOut=obj2d->_shape->getTheta();
}

void
Object2D_setX(Object2D * obj2d,
              double x)
{
obj2d->_shape->move(x,obj2d->_shape->getY(),obj2d->_shape->getTheta());
}

double
Object2D_getX(const Object2D * obj2d)
{
return(obj2d->_shape->getX());
}

void
Object2D_setY(Object2D * obj2d,
              double y)
{
obj2d->_shape->move(obj2d->_shape->getX(),y,obj2d->_shape->getTheta());
}

double
Object2D_getY(const Object2D * obj2d)
{
return(obj2d->_shape->getY());
}

void
Object2D_setTheta(Object2D * obj2d,
                  double theta)
{
obj2d->_shape->move(obj2d->_shape->getX(),obj2d->_shape->getY(),theta);
}

double
Object2D_getTheta(const Object2D * obj2d)
{
return(obj2d->_shape->getTheta());
}

/*-------- Motion ---------------------------------------------------------*/

void
Object2D_translate(Object2D * obj2d,
                   double dx,
                   double dy)
{
double theta=obj2d->_shape->getTheta();
double c=cos(theta);
double s=sin(theta);
obj2d->_shape->move(obj2d->_shape->getX()+dx*c-dy*s,
                    obj2d->_shape->getY()+dx*s+dy*c,theta);
}

void
Object2D_rotate(Object2D * obj2d,
                double dTheta)
{
double theta=obj2d->_shape->getTheta()+dTheta;
while(theta>M_PI) theta-=2*M_PI;
while(theta<=-M_PI) theta+=2*M_PI;
obj2d->_shape->move(obj2d->_shape->getX(),obj2d->_shape->getY(),theta);
}


/*-------- Detection -------------------------------------------------------*/

int /* 0: outside    !=0: inside */
Object2D_isInside(const Object2D * obj2d,
                  double x,
                  double y)
{
return(obj2d->_shape->isInside(x,y) ? 1 : 0);
}

int /* 0: no intersection found   !=0: intersection found */
Object2D_intersectRay(const Object2D * obj2d,
                      double xRay,
                      double yRay,
                      double thetaRay,
                      double * xOut,
                      double * yOut)
{
double xDir=cos(thetaRay);
double yDir=sin(thetaRay);
return(obj2d->_shape->intersectRay(xRay,yRay,xDir,yDir,*xOut,*yOut) ? 1 : 0);
}

/*-------- Transformation --------------------------------------------------*/

void
Object2D_globalToLocalPosition(const Object2D * obj2d,
                               double * xInOut,
                               double * yInOut)
{
double x=*xInOut-obj2d->_shape->getX();
double y=*yInOut-obj2d->_shape->getY();
double d=sqrt(x*x+y*y);
double a=(d ? atan2(y,x) : 0.0);
a-=obj2d->_shape->getTheta();
*xInOut=d*cos(a);
*yInOut=d*sin(a);
}

void
Object2D_localToGlobalPosition(const Object2D * obj2d,
                               double * xInOut,
                               double * yInOut)
{
double x=*xInOut;
double y=*yInOut;
double d=sqrt(x*x+y*y);
double a=(d ? atan2(y,x) : 0.0);
a+=obj2d->_shape->getTheta();
*xInOut=obj2d->_shape->getX()+d*cos(a);
*yInOut=obj2d->_shape->getY()+d*sin(a);
}

double
Object2D_globalToLocalOrientation(const Object2D * obj2d,
                                  double orientation)
{
orientation-=obj2d->_shape->getTheta();
while(orientation>M_PI) orientation-=2*M_PI;
while(orientation<=-M_PI) orientation+=2*M_PI;
return(orientation);
}

double
Object2D_localToGlobalOrientation(const Object2D * obj2d,
                                  double orientation)
{
orientation+=obj2d->_shape->getTheta();
while(orientation>M_PI) orientation-=2*M_PI;
while(orientation<=-M_PI) orientation+=2*M_PI;
return(orientation);
}

/*-------- Representation --------------------------------------------------*/

void
Object2D_setColor(Object2D * obj2d,
                  const char * colorName)
{
obj2d->_shape->setColor(colorName);
}

const char *
Object2D_getColor(const Object2D * obj2d)
{
return(obj2d->_shape->getColor());
}

void
Object2D_setLayer(Object2D * obj2d,
                  int layer)
{
bool select=_viewer->isSelected(obj2d->_shape);
obj2d->_shape->setLayer(layer);
if(select) _viewer->select(obj2d->_shape, true);
}
                  
int
Object2D_getLayer(const Object2D * obj2d)
{
return(obj2d->_shape->getLayer());
}

static
void
_changeShape(Object2D * obj2d,
             GuiShape2D * sh)
{
sh->setLayer(obj2d->_shape->getLayer());
sh->setKeyCallback(_keyCB,obj2d);
sh->setMotionCallback(_motionCB,obj2d);
if(_viewer->isSelected(obj2d->_shape)) _viewer->select(sh,true);
delete obj2d->_shape;
obj2d->_shape=sh;
obj2d->_isImage=false;
}

void
Object2D_point(Object2D * obj2d)
{
_changeShape(obj2d,
             new GuiPoint(obj2d->_shape->getX(),
                          obj2d->_shape->getY(),
                          obj2d->_shape->getTheta(),
                          obj2d->_shape->getColor()));
}

void
Object2D_text(Object2D* obj2d,
              const char * text)
{
_changeShape(obj2d,
             new GuiText(obj2d->_shape->getX(),
                         obj2d->_shape->getY(),
                         obj2d->_shape->getTheta(),
                         obj2d->_shape->getColor(),
                         text));
}

void
Object2D_line(Object2D* obj2d,
              double length)
{
_changeShape(obj2d,
             new GuiLine(obj2d->_shape->getX(),
                         obj2d->_shape->getY(),
                         obj2d->_shape->getTheta(),
                         obj2d->_shape->getColor(),
                         length));
}

void
Object2D_square(Object2D * obj2d,
                double side,
                int filled)
{
_changeShape(obj2d,
             new GuiSquare(obj2d->_shape->getX(),
                           obj2d->_shape->getY(),
                           obj2d->_shape->getTheta(),
                           obj2d->_shape->getColor(),
                           side,
                           filled));
}

void
Object2D_rectangle(Object2D * obj2d,
                   double length,
                   double width,
                   int filled)
{
_changeShape(obj2d,
             new GuiRectangle(obj2d->_shape->getX(),
                              obj2d->_shape->getY(),
                              obj2d->_shape->getTheta(),
                              obj2d->_shape->getColor(),
                              length,
                              width,
                              filled));
}

void
Object2D_polyline(Object2D * obj2d,
                  unsigned int nbPoints,
                  const double * xPoints,
                  const double * yPoints)
{
_changeShape(obj2d,
             new GuiPolyline(obj2d->_shape->getX(),
                             obj2d->_shape->getY(),
                             obj2d->_shape->getTheta(),
                             obj2d->_shape->getColor(),
                             nbPoints,
                             xPoints,
                             yPoints));
}

void
Object2D_polygon(Object2D * obj2d,
                 unsigned int nbPoints,
                 const double * xPoints,
                 const double * yPoints,
                 int filled)
{
_changeShape(obj2d,
             new GuiPolygon(obj2d->_shape->getX(),
                            obj2d->_shape->getY(),
                            obj2d->_shape->getTheta(),
                            obj2d->_shape->getColor(),
                            nbPoints,
                            xPoints,
                            yPoints,
                            filled));
}

void
Object2D_circle(Object2D * obj2d,
                double radius,
                int filled)
{
_changeShape(obj2d,
             new GuiCircle(obj2d->_shape->getX(),
                           obj2d->_shape->getY(),
                           obj2d->_shape->getTheta(),
                           obj2d->_shape->getColor(),
                           radius,
                           filled));
}

int /* 0: failure    !=0: success */
Object2D_image(Object2D * obj2d,
               const char * fileName,
               double pixelScale)
{
GuiImage * sh=new GuiImage(obj2d->_shape->getX(),
                           obj2d->_shape->getY(),
                           obj2d->_shape->getTheta(),
                           obj2d->_shape->getColor(),
                           fileName,
                           (const StringsVect *)0,
                           pixelScale);
if(sh->fail())
  {
  delete sh;
  return(0);
  }
else
  {
  _changeShape(obj2d,sh);
  obj2d->_isImage=true;
  return(1);
  }
}

int /* 0: failure    !=0: success */
Object2D_getImagePixelAt(Object2D * obj2d,
                         double x,
                         double y,
                         int * redOut,
                         int * greenOut,
                         int * blueOut)
{
if(!obj2d->_isImage)
  {
  *redOut=-1;
  *greenOut=-1;
  *blueOut=-1;
  return(0);
  }
GuiImage * sh=(GuiImage *)obj2d->_shape;
double xx=x-sh->getX();
double yy=y-sh->getY();
double sinTh=sin(sh->getTheta());
double cosTh=cos(sh->getTheta());
int xi=(int)::floor(sh->getNbColumns()*0.5+(xx*cosTh+yy*sinTh)/sh->getScale());
int yi=(int)::floor(sh->getNbLines()*0.5+(yy*cosTh-xx*sinTh)/sh->getScale());
if((xi>=0)&&(xi<sh->getNbColumns())&&
   (yi>=0)&&(yi<sh->getNbLines()))
  {
  unsigned char r,g,b;
  sh->getRGB(sh->getPixel(xi,yi),r,g,b);
  *redOut=r;
  *greenOut=g;
  *blueOut=b;
  return(1);
  }
else
  {
  *redOut=-1;
  *greenOut=-1;
  *blueOut=-1;
  return(0);
  }
}

int /* >=0: pixel number    <0: failure */
Object2D_getImagePixelNumberAt(Object2D * obj2d,
                               double x,
                               double y)
{
if(!obj2d->_isImage)
  {
  return(-1);
  }
GuiImage * sh=(GuiImage *)obj2d->_shape;
double xx=x-sh->getX();
double yy=y-sh->getY();
double sinTh=sin(sh->getTheta());
double cosTh=cos(sh->getTheta());
int xi=(int)::floor(sh->getNbColumns()*0.5+(xx*cosTh+yy*sinTh)/sh->getScale());
int yi=(int)::floor(sh->getNbLines()*0.5+(yy*cosTh-xx*sinTh)/sh->getScale());
if((xi>=0)&&(xi<sh->getNbColumns())&&
   (yi>=0)&&(yi<sh->getNbLines()))
  {
  return(sh->getPixel(xi,yi));
  }
else
  {
  return(-1);
  }
}

int /* 0: failure    !=0: success */
Object2D_setImagePixelNumberAt(Object2D * obj2d,
		               int pixel,
                               double x,
                               double y)
{
if(!obj2d->_isImage)
  {
  return(0);
  }
GuiImage * sh=(GuiImage *)obj2d->_shape;
if((pixel<0)||(pixel>=sh->getNbColors()))
  {
  return(0);
  }
double xx=x-sh->getX();
double yy=y-sh->getY();
double sinTh=sin(sh->getTheta());
double cosTh=cos(sh->getTheta());
int xi=(int)::floor(sh->getNbColumns()*0.5+(xx*cosTh+yy*sinTh)/sh->getScale());
int yi=(int)::floor(sh->getNbLines()*0.5+(yy*cosTh-xx*sinTh)/sh->getScale());
if((xi>=0)&&(xi<sh->getNbColumns())&&
   (yi>=0)&&(yi<sh->getNbLines()))
  {
  sh->setPixel(xi,yi,pixel);
  return(1);
  }
else
  {
  return(0);
  }
}

int /* number of colors */
Object2D_getImageNbColors(Object2D * obj2d)
{
if(!obj2d->_isImage)
  {
  return(0);
  }
GuiImage * sh=(GuiImage *)obj2d->_shape;
return(sh->getNbColors());
}

int /* 0: failure    !=0: success */
Object2D_getImageRGB(Object2D * obj2d,
		     int pixel,
                     int * redOut,
                     int * greenOut,
                     int * blueOut)
{
if(!obj2d->_isImage)
  {
  *redOut=-1;
  *greenOut=-1;
  *blueOut=-1;
  return(0);
  }
GuiImage * sh=(GuiImage *)obj2d->_shape;
if((pixel<0)||(pixel>=sh->getNbColors()))
  {
  *redOut=-1;
  *greenOut=-1;
  *blueOut=-1;
  return(0);
  }
unsigned char r,g,b;
sh->getRGB(pixel,r,g,b);
*redOut=r;
*greenOut=g;
*blueOut=b;
return(1);
}

/*-------- Graphical application template ----------------------------------*/

void
graphic_init(const char * windowName,
             const char * fontName)
{
setenv("ORISFONT",fontName,1);
GuiManager * manager=getGuiManager();
if(!manager) manager=new GuiManager();
if(manager->fail())
  {
  cerr << "Can't connect to X Server" << endl;
  delete manager;
  exit(1);
  }
if(_viewer) delete _viewer;
if(_mainWin) delete _mainWin;
_mainWin = new GuiWindow(windowName);
_mainWin->setInWidth(200);
_mainWin->setInHeight(200);
_viewer = new GuiViewer2D(_mainWin);
_mainWin->setCallback(_viewerResizeCB,(void *)_viewer);
_mainWin->setCloseCallback(_viewerCloseCB,(void *)0);
getGuiManager()->show();
}

void
graphic_setWidth(int width)
{
if(_mainWin) _mainWin->setInWidth(width);
}

void
graphic_setHeight(int height)
{
if(_mainWin) _mainWin->setInHeight(height);
}

void
graphic_setBackground(const char * colorName)
{
if(_viewer) _viewer->setBackground(colorName);
}

void graphic_setViewPoint(double x,
		          double y,
			  double scale)
{
if(_viewer) _viewer->setViewPoint(x,y,scale);
}

void graphic_getViewPoint(double * xOut,
		          double * yOut,
			  double * scaleOut)
{
if(_viewer) _viewer->getViewPoint(*xOut,*yOut,*scaleOut);
}

void
graphic_autoscale(void)
{
if(_viewer) _viewer->autoScale();
}

void
graphic_run(void * userData)
{
_userData=userData;
GuiManager * manager = getGuiManager();
if(manager)
  {
  if(_viewer) _viewer->setKeyCallback(_viewerKeyCB,(void *)0);
  manager->setTask(graphic_mainLoop,_userData);
  manager->mainLoop();
  }
}

void
graphic_reset(void)
{
if(_viewer)
  {
  delete _viewer;
  _viewer=(GuiViewer2D *)0;
  }
if(_mainWin)
  {
  delete _mainWin;
  _mainWin=(GuiWindow *)0;
  }
GuiManager * manager=getGuiManager();
if(manager) delete manager;
}

/*--------------------------------------------------------------------------*/
